@@ -2,9 +2,9 @@
 # -*- coding: utf-8 -*-
 
 """
-Copyright (C) 2009-2012 Wolfgang Rohdewald <wolfgang@rohdewald.de>
+Copyright (C) 2009-2014 Wolfgang Rohdewald <wolfgang@rohdewald.de>
 
-kajongg is free software you can redistribute it and/or modify
+Kajongg is free software you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation either version 2 of the License, or
 (at your option) any later version.
@@ -21,11 +21,12 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
 import traceback
 import datetime
+import weakref
 
 from twisted.spread import pb
 from twisted.internet.defer import Deferred
 
-from util import m18nE, logInfo, logDebug, logException
+from log import m18nE, logInfo, logDebug, logException
 from message import Message
 from common import Debug
 from move import Move
@@ -33,14 +34,39 @@ from move import Move
 class Request(object):
     """holds a Deferred and related attributes, used as part of a DeferredBlock"""
     def __init__(self, block, deferred, user, about):
-        self.block = block
+        self._block = weakref.ref(block)
         self.deferred = deferred
-        self.user = user
-        self.about = about
+        self._user = weakref.ref(user)
+        self._about = weakref.ref(about) if about else None
         self.answer = None
         self.args = None
         self.startTime = datetime.datetime.now()
-        self.player = self.block.playerForUser(user)
+        player = self.block.playerForUser(user)
+        self._player = weakref.ref(player) if player else None
+
+    @property
+    def block(self):
+        """hide weakref"""
+        if self._block:
+            return self._block()
+
+    @property
+    def user(self):
+        """hide weakref"""
+        if self._user:
+            return self._user()
+
+    @property
+    def about(self):
+        """hide weakref"""
+        if self._about:
+            return self._about()
+
+    @property
+    def player(self):
+        """hide weakref"""
+        if self._player:
+            return self._player()
 
     def gotAnswer(self, rawAnswer):
         """convert the wired answer into something more useful"""
@@ -69,8 +95,11 @@ class Request(object):
             answer = str(self.answer)
         else:
             answer = 'OPEN'
-        result = '[{id:>4}] {cmd}->{cls}({receiver:<10}): {answer}'.format(
-            cls=self.user.__class__.__name__, id=id(self)%10000, cmd=cmd, receiver=self.user.name,
+        result = ''
+        if Debug.deferredBlock:
+            result += '[{id:>4}] '.format(id=id(self)%10000)
+        result += '{cmd}->{cls}({receiver:<10}): {answer}'.format(
+            cls=self.user.__class__.__name__, cmd=cmd, receiver=self.user.name,
             answer=answer)
         if self.age():
             result += ' after {} sec'.format(self.age())
@@ -91,9 +120,12 @@ class Request(object):
 
     def pretty(self):
         """for debug output"""
-        result = '[{id:>4}] {cmd:<12}<-{cls:>6}({receiver:<10}): ANS={answer}'.format(
+        result = ''
+        if Debug.deferredBlock:
+            result += '[{id:>4}] '.format(id=id(self)%10000)
+        result += '{cmd:<12}<-{cls:>6}({receiver:<10}): ANS={answer}'.format(
             cls=self.user.__class__.__name__,
-            id=id(self)%10000, answer=self.prettyAnswer(), cmd=self.deferred.command, receiver=self.user.name)
+            answer=self.prettyAnswer(), cmd=self.deferred.command, receiver=self.user.name)
         if self.age() > 0:
             result += ' after {} sec'.format(self.age())
         return result
@@ -127,12 +159,6 @@ class DeferredBlock(object):
                     for block in DeferredBlock.blocks:
                         logInfo(str(block))
 
-    def __cleanup(self):
-        """must do this for Request objects to be freeable"""
-        for request in self.requests:
-            request.block = None    # break reference cycle
-            del request
-
     def debugPrefix(self, marker=''):
         """prefix for debug message"""
         return 'T{table} B[{id:>4}] {caller:<15} {marker:<3}(out={out})'.format(
@@ -256,7 +282,6 @@ class DeferredBlock(object):
     def callbackIfDone(self):
         """if we are done, convert received answers to something more useful and callback"""
         if self.completed:
-            self.__cleanup()
             return
         assert self.outstanding >= 0, 'callbackIfDone: outstanding %d' % self.outstanding
         if self.outstanding == 0 and self.callbackMethod is not None:
@@ -270,8 +295,8 @@ class DeferredBlock(object):
                     answerList = []
                     for answer in set(x.prettyAnswer() for x in self.requests if x.deferred.command == command):
                         answerList.append((answer, list(x for x in self.requests
-                            if x.deferred.command == command and answer==x.prettyAnswer())))
-                    answerList = sorted(answerList, key=lambda x:len(x[1]))
+                            if x.deferred.command == command and answer == x.prettyAnswer())))
+                    answerList = sorted(answerList, key=lambda x: len(x[1]))
                     answerTexts = []
                     if len(answerList) == 1:
                         answerTexts.append('{answer} from all'.format(answer=answerList[-1][0]))
@@ -288,7 +313,6 @@ class DeferredBlock(object):
                 self.debug('END', '{answers} {method}'.format(method=methodName, answers=' / '.join(commandText)))
             if self.callbackMethod is not False:
                 self.callbackMethod(self.requests, *self.__callbackArgs)
-            self.__cleanup()
 
     def prettyCallback(self):
         """pretty string for callbackMethod"""
@@ -320,7 +344,7 @@ class DeferredBlock(object):
             kwargs['score'] = str(about.hand)
         if game and game.gameid and 'token' not in kwargs:
             # this lets the client assert that the message is meant for the current hand
-            kwargs['token'] = game.handId()
+            kwargs['token'] = game.handId.token()
         else:
             kwargs['token'] = None
 
